##############################################################################
#
# Copyright (c) Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest


from persistent.tests.utils import TrivialJar
from persistent.tests.utils import copy_test

# pylint:disable=blacklisted-name, protected-access

class Test_default(unittest.TestCase):

    def _getTargetClass(self):
        from persistent.mapping import default
        return default

    def _makeOne(self, func):
        return self._getTargetClass()(func)

    def test___get___from_class(self):
        def _test(inst):
            raise AssertionError("Must not be caled")

        descr = self._makeOne(_test)
        class Foo(object):
            testing = descr
        self.assertIs(Foo.testing, descr)


    def test___get___from_instance(self):
        _called_with = []
        def _test(inst):
            _called_with.append(inst)
            return 'TESTING'
        descr = self._makeOne(_test)
        class Foo(object):
            testing = descr
        foo = Foo()
        self.assertEqual(foo.testing, 'TESTING')
        self.assertEqual(_called_with, [foo])


class PersistentMappingTests(unittest.TestCase):

    def _getTargetClass(self):
        from persistent.mapping import PersistentMapping
        return PersistentMapping

    def _makeJar(self):
        return TrivialJar()

    def _makeOne(self, *args, **kwargs):
        inst = self._getTargetClass()(*args, **kwargs)
        inst._p_jar = self._makeJar()
        return inst

    def test_volatile_attributes_not_persisted(self):
        # http://www.zope.org/Collectors/Zope/2052
        m = self._makeOne()
        m.foo = 'bar'
        m._v_baz = 'qux'
        state = m.__getstate__()
        self.assertTrue('foo' in state)
        self.assertFalse('_v_baz' in state)

    def testTheWorld(self):
        # Test constructors
        l0 = {}
        l1 = {0:0}
        l2 = {0:0, 1:1}
        u = self._makeOne()
        u0 = self._makeOne(l0)
        u1 = self._makeOne(l1)
        u2 = self._makeOne(l2)

        uu = self._makeOne(u)
        uu0 = self._makeOne(u0)
        uu1 = self._makeOne(u1)
        uu2 = self._makeOne(u2)

        class OtherMapping(dict):
            def __init__(self, initmapping):
                dict.__init__(self)
                self.__data = initmapping
            def items(self):
                raise AssertionError("Not called")
        self._makeOne(OtherMapping(u0))
        self._makeOne([(0, 0), (1, 1)])

        # Test __repr__
        eq = self.assertEqual

        eq(str(u0), str(l0), "str(u0) == str(l0)")
        eq(repr(u1), repr(l1), "repr(u1) == repr(l1)")

        # Test __cmp__ and __len__
        try:
            cmp
        except NameError:
            def cmp(a, b):
                if a == b:
                    return 0
                if hasattr(a, 'items'):
                    a = sorted(a.items())
                    b = sorted(b.items())
                if a < b:
                    return -1
                return 1

        def mycmp(a, b):
            r = cmp(a, b)
            if r < 0:
                return -1
            if r > 0:
                return 1
            return r

        to_test = [l0, l1, l2, u, u0, u1, u2, uu, uu0, uu1, uu2]
        for a in to_test:
            for b in to_test:
                eq(mycmp(a, b), mycmp(len(a), len(b)),
                   "mycmp(a, b) == mycmp(len(a), len(b))")

        # Test __getitem__

        for i, val in enumerate(u2):
            eq(val, i, "u2[i] == i")

        # Test get

        for i in range(len(u2)):
            eq(u2.get(i), i, "u2.get(i) == i")
            eq(u2.get(i, 5), i, "u2.get(i, 5) == i")

        for i in min(u2)-1, max(u2)+1:
            eq(u2.get(i), None, "u2.get(i) == None")
            eq(u2.get(i, 5), 5, "u2.get(i, 5) == 5")

        # Test __setitem__

        uu2[0] = 0
        uu2[1] = 100
        uu2[2] = 200

        # Test __delitem__

        del uu2[1]
        del uu2[0]
        with self.assertRaises(KeyError):
            del uu2[0]

        # Test __contains__
        for i in u2:
            self.assertTrue(i in u2, "i in u2")
        for i in min(u2)-1, max(u2)+1:
            self.assertTrue(i not in u2, "i not in u2")

        # Test update

        l = {"a":"b"}
        u = self._makeOne(l)
        u.update(u2)
        for i in u:
            self.assertTrue(i in l or i in u2, "i in l or i in u2")
        for i in l:
            self.assertTrue(i in u, "i in u")
        for i in u2:
            self.assertTrue(i in u, "i in u")

        # Test setdefault

        x = u2.setdefault(0, 5)
        eq(x, 0, "u2.setdefault(0, 5) == 0")

        x = u2.setdefault(5, 5)
        eq(x, 5, "u2.setdefault(5, 5) == 5")
        self.assertTrue(5 in u2, "5 in u2")

        # Test pop

        x = u2.pop(1)
        eq(x, 1, "u2.pop(1) == 1")
        self.assertTrue(1 not in u2, "1 not in u2")

        with self.assertRaises(KeyError):
            u2.pop(1)

        x = u2.pop(1, 7)
        eq(x, 7, "u2.pop(1, 7) == 7")

        # Test popitem

        items = list(u2.items())
        key, value = u2.popitem()
        self.assertTrue((key, value) in items, "key, value in items")
        self.assertTrue(key not in u2, "key not in u2")

        # Test clear

        u2.clear()
        eq(u2, {}, "u2 == {}")

    def test___repr___converts_legacy_container_attr(self):
        # In the past, PM used a _container attribute. For some time, the
        # implementation continued to use a _container attribute in pickles
        # (__get/setstate__) to be compatible with older releases.  This isn't
        # really necessary any more. In fact, releases for which this might
        # matter can no longer share databases with current releases.  Because
        # releases as recent as 3.9.0b5 still use _container in saved state, we
        # need to accept such state, but we stop producing it.
        pm = self._makeOne()
        self.assertEqual(pm.__dict__, {'data': {}})
        # Make it look like an older instance
        pm.__dict__.clear()
        pm.__dict__['_container'] = {'a': 1}
        self.assertEqual(pm.__dict__, {'_container': {'a': 1}})
        pm._p_changed = 0
        self.assertEqual(repr(pm), "{'a': 1}")
        self.assertEqual(pm.__dict__, {'data': {'a': 1}})
        self.assertEqual(pm.__getstate__(), {'data': {'a': 1}})

    def test_update_keywords(self):
        # Prior to https://github.com/zopefoundation/persistent/issues/126,
        # PersistentMapping didn't accept keyword arguments to update as
        # the builtin dict and the UserDict do.
        # Here we make sure it does. We use some names that have been
        # seen to be special in signatures as well to make sure that
        # we don't interpret them incorrectly.
        pm = self._makeOne()
        # Our older implementation was ``def update(self, b)``, so ``b``
        # is potentially a keyword argument in the wild; the behaviour in that
        # corner case has changed.
        pm.update(b={'a': 42})
        self.assertEqual(pm, {'b': {'a': 42}})

        pm = self._makeOne()
        # Our previous implementation would explode with a TypeError
        pm.update(b=42)
        self.assertEqual(pm, {'b': 42})

        pm = self._makeOne()
        # ``other`` shows up in a Python 3 signature.
        pm.update(other=42)
        self.assertEqual(pm, {'other': 42})
        pm = self._makeOne()
        pm.update(other={'a': 42})
        self.assertEqual(pm, {'other': {'a': 42}})

        pm = self._makeOne()
        pm.update(a=1, b=2)
        self.assertEqual(pm, {'a': 1, 'b': 2})

    def test_clear_nonempty(self):
        pm = self._makeOne({'a': 42})
        self.assertFalse(pm._p_changed)
        pm.clear()
        self.assertTrue(pm._p_changed)

    def test_clear_empty(self):
        pm = self._makeOne()
        self.assertFalse(pm._p_changed)
        pm.clear()
        self.assertFalse(pm._p_changed)

    def test_clear_no_jar(self):
        # https://github.com/zopefoundation/persistent/issues/139
        self._makeOne = self._getTargetClass()
        self.test_clear_empty()

        pm = self._makeOne(a=42)
        pm.clear()
        self.assertFalse(pm._p_changed)

    def test_clear_empty_legacy_container(self):
        pm = self._makeOne()
        pm.__dict__['_container'] = pm.__dict__.pop('data')
        self.assertFalse(pm._p_changed)
        pm.clear()
        # Migration happened
        self.assertIn('data', pm.__dict__)
        # and we are marked as changed.
        self.assertTrue(pm._p_changed)

    def test_copy(self):
        pm = self._makeOne()
        pm['key'] = 42
        copy = copy_test(self, pm)
        self.assertEqual(42, copy['key'])

    def test_copy_legacy_container(self):
        pm = self._makeOne()
        pm['key'] = 42
        pm.__dict__['_container'] = pm.__dict__.pop('data')

        self.assertNotIn('data', pm.__dict__)
        self.assertIn('_container', pm.__dict__)

        copy = copy_test(self, pm)
        self.assertNotIn('_container', copy.__dict__)
        self.assertIn('data', copy.__dict__)
        self.assertEqual(42, copy['key'])


class Test_legacy_PersistentDict(unittest.TestCase):

    def _getTargetClass(self):
        from persistent.dict import PersistentDict
        return PersistentDict

    def test_PD_is_alias_to_PM(self):
        from persistent.mapping import PersistentMapping
        self.assertIs(self._getTargetClass(), PersistentMapping)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
__all__ = ('TimeStamp',)

import datetime
import math
import struct
import sys

from persistent._compat import use_c_impl

_RAWTYPE = bytes
_MAXINT = sys.maxsize

_ZERO = b'\x00' * 8

__all__ = [
    'TimeStamp',
    'TimeStampPy',
]

try:
    # Make sure to overflow and wraparound just
    # like the C code does.
    from ctypes import c_long
except ImportError: # pragma: no cover
    # XXX: This is broken on 64-bit windows, where
    # sizeof(long) != sizeof(Py_ssize_t)
    # sizeof(long) == 4, sizeof(Py_ssize_t) == 8
    # It can be fixed by setting _MAXINT = 2 ** 31 - 1 on all
    # win32 platforms, but then that breaks PyPy3 64 bit for an unknown
    # reason.
    c_long = None
    def _wraparound(x):
        return int(((x + (_MAXINT + 1)) & ((_MAXINT << 1) + 1)) - (_MAXINT + 1))
else:
    def _wraparound(x):
        return c_long(x).value


class _UTCClass(datetime.tzinfo):
    # A Python 2 implementation of a UTC tzinfo.
    def tzname(self, dt):
        return 'UTC'
    def utcoffset(self, dt):
        return datetime.timedelta(0, 0, 0)
    def dst(self, dt):
        return None
    def fromutc(self, dt):
        return dt

try:
    from datetime import timezone
    def _UTC():
        return timezone.utc
except ImportError: # pragma: no cover
    # Python 2
    def _UTC(_inst=_UTCClass()):
        return _inst


def _makeUTC(y, mo, d, h, mi, s):
    s = round(s, 6) # microsecond precision, to match the C implementation
    usec, sec = math.modf(s)
    sec = int(sec)
    usec = int(usec * 1e6)
    return datetime.datetime(y, mo, d, h, mi, sec, usec, tzinfo=_UTC())

_EPOCH = _makeUTC(1970, 1, 1, 0, 0, 0)

_TS_SECOND_BYTES_BIAS = 60.0 / (1<<16) / (1<<16)

def _makeRaw(year, month, day, hour, minute, second):
    a = (((year - 1900) * 12 + month - 1) * 31 + day - 1)
    a = (a * 24 + hour) * 60 + minute
    b = int(second / _TS_SECOND_BYTES_BIAS) # Don't round() this; the C version just truncates
    return struct.pack('>II', a, b)

def _parseRaw(octets):
    a, b = struct.unpack('>II', octets)
    minute = a % 60
    hour = a // 60 % 24
    day = a // (60 * 24) % 31 + 1
    month = a // (60 * 24 * 31) % 12 + 1
    year = a // (60 * 24 * 31 * 12) + 1900
    second = b * _TS_SECOND_BYTES_BIAS
    return (year, month, day, hour, minute, second)



@use_c_impl
class TimeStamp(object):
    __slots__ = ('_raw', '_elements')

    def __init__(self, *args):
        self._elements = None
        if len(args) == 1:
            raw = args[0]
            if not isinstance(raw, _RAWTYPE):
                raise TypeError('Raw octets must be of type: %s' % _RAWTYPE)
            if len(raw) != 8:
                raise TypeError('Raw must be 8 octets')
            self._raw = raw
        elif len(args) == 6:
            self._raw = _makeRaw(*args) # pylint:disable=no-value-for-parameter
            # Note that we don't preserve the incoming arguments in self._elements,
            # we derive them from the raw value. This is because the incoming
            # seconds value could have more precision than would survive
            # in the raw data, so we must be consistent.
        else:
            raise TypeError('Pass either a single 8-octet arg '
                            'or 5 integers and a float')

        self._elements = _parseRaw(self._raw)

    def raw(self):
        return self._raw

    def __repr__(self):
        return repr(self._raw)

    def __str__(self):
        return "%4.4d-%2.2d-%2.2d %2.2d:%2.2d:%09.6f" % (
            self.year(), self.month(), self.day(),
            self.hour(), self.minute(),
            self.second())

    def year(self):
        return self._elements[0]

    def month(self):
        return self._elements[1]

    def day(self):
        return self._elements[2]

    def hour(self):
        return self._elements[3]

    def minute(self):
        return self._elements[4]

    def second(self):
        return self._elements[5]

    def timeTime(self):
        """ -> seconds since epoch, as a float.
        """
        delta = _makeUTC(*self._elements) - _EPOCH
        return delta.days * 86400 + delta.seconds + delta.microseconds / 1e6

    def laterThan(self, other):
        """ Return a timestamp instance which is later than 'other'.

        If self already qualifies, return self.

        Otherwise, return a new instance one moment later than 'other'.
        """
        if not isinstance(other, self.__class__):
            raise ValueError()
        # pylint:disable=protected-access
        if self._raw > other._raw:
            return self
        a, b = struct.unpack('>II', other._raw)
        later = struct.pack('>II', a, b + 1)
        return self.__class__(later)

    def __eq__(self, other):
        try:
            return self.raw() == other.raw()
        except AttributeError:
            return NotImplemented

    def __ne__(self, other):
        try:
            return self.raw() != other.raw()
        except AttributeError:
            return NotImplemented

    def __hash__(self):
        # Match the C implementation
        a = bytearray(self._raw)
        x = a[0] << 7
        for i in a:
            x = (1000003 * x) ^ i
        x ^= 8

        x = _wraparound(x)

        if x == -1: # pragma: no cover
            # The C version has this condition, but it's not clear
            # why; it's also not immediately obvious what bytestring
            # would generate this---hence the no-cover
            x = -2
        return x

    # Now the rest of the comparison operators
    # Sigh. Python 2.6 doesn't have functools.total_ordering
    # so we have to do it by hand
    def __lt__(self, other):
        try:
            return self.raw() < other.raw()
        except AttributeError:
            return NotImplemented

    def __gt__(self, other):
        try:
            return self.raw() > other.raw()
        except AttributeError:
            return NotImplemented

    def __le__(self, other):
        try:
            return self.raw() <= other.raw()
        except AttributeError:
            return NotImplemented

    def __ge__(self, other):
        try:
            return self.raw() >= other.raw()
        except AttributeError:
            return NotImplemented


# This name is bound by the ``@use_c_impl`` decorator to the class defined above.
# We make sure and list it statically, though, to help out linters.
TimeStampPy = TimeStampPy # pylint:disable=undefined-variable,self-assigning-variable




class TrivialJar(object):
    """
    Jar that only supports registering objects so ``_p_changed``
    can be tested.
    """

    def register(self, ob):
        """Does nothing"""

class ResettingJar(object):
    """Testing stub for _p_jar attribute.
    """
    def __init__(self):
        from persistent import PickleCache # XXX stub it!
        from persistent.interfaces import IPersistentDataManager
        from zope.interface import directlyProvides
        self.cache = self._cache = PickleCache(self)
        self.oid = 1
        self.registered = {}
        directlyProvides(self, IPersistentDataManager)

    def add(self, obj):
        import struct
        obj._p_oid = struct.pack(">Q", self.oid)
        self.oid += 1
        obj._p_jar = self
        self.cache[obj._p_oid] = obj


    # the following methods must be implemented to be a jar

    def setstate(self, obj):
        # Trivial setstate() implementation that just re-initializes
        # the object.  This isn't what setstate() is supposed to do,
        # but it suffices for the tests.
        obj.__class__.__init__(obj)

class RememberingJar(object):
    """Testing stub for _p_jar attribute.
    """
    def __init__(self):
        from persistent import PickleCache # XXX stub it!
        self.cache = PickleCache(self)
        self.oid = 1
        self.registered = {}

    def add(self, obj):
        import struct
        obj._p_oid = struct.pack(">Q", self.oid)
        self.oid += 1
        obj._p_jar = self
        self.cache[obj._p_oid] = obj
        # Remember object's state for later.
        self.obj = obj
        self.remembered = obj.__getstate__()


    def fake_commit(self):
        self.remembered = self.obj.__getstate__()
        self.obj._p_changed = 0

    # the following methods must be implemented to be a jar

    def register(self, obj):
        self.registered[obj] = 1

    def setstate(self, obj):
        # Trivial setstate() implementation that resets the object's
        # state as of the time it was added to the jar.
        # This isn't what setstate() is supposed to do,
        # but it suffices for the tests.
        obj.__setstate__(self.remembered)


def copy_test(self, obj):
    import copy
    # Test copy.copy. Do this first, because depending on the
    # version of Python, `UserDict.copy()` may wind up
    # mutating the original object's ``data`` (due to our
    # BWC with ``_container``). This shows up only as a failure
    # of coverage.
    obj.test = [1234]  # Make sure instance vars are also copied.
    obj_copy = copy.copy(obj)
    self.assertIsNot(obj.data, obj_copy.data)
    self.assertEqual(obj.data, obj_copy.data)
    self.assertIs(obj.test, obj_copy.test)

    # Test internal copy
    obj_copy = obj.copy()
    self.assertIsNot(obj.data, obj_copy.data)
    self.assertEqual(obj.data, obj_copy.data)

    return obj_copy


def skipIfNoCExtension(o):
    import unittest
    from persistent._compat import _should_attempt_c_optimizations
    from persistent._compat import _c_optimizations_available
    from persistent._compat import _c_optimizations_ignored

    if _should_attempt_c_optimizations() and not _c_optimizations_available(): # pragma: no cover
        return unittest.expectedFailure(o)
    return unittest.skipIf(
        _c_optimizations_ignored() or not _c_optimizations_available(),
        "The C extension is not available"
    )(o)
